import numpy as np

from krisi import score
from krisi.evaluate.library.default_metrics_regression import (
    all_regression_metrics,
    low_computation_regression_mterics,
    minimal_regression_metrics,
)

score(
    y=np.random.normal(0, 0.1, 1000),
    predictions=np.random.normal(0, 0.1, 1000),
    default_metrics=all_regression_metrics,  # This is the default
).print_summary()

score(
    y=np.random.normal(0, 0.1, 1000),
    predictions=np.random.normal(0, 0.1, 1000),
    default_metrics=minimal_regression_metrics,
).print_summary(input_analysis=False)

score(
    y=np.random.normal(0, 0.1, 1000),
    predictions=np.random.normal(0, 0.1, 1000),
    default_metrics=low_computation_regression_mterics,
).print_summary(input_analysis=False)



from typing import List

import numpy as np
import pandas as pd
from sklearn.metrics import (
    mean_absolute_error,
    mean_absolute_percentage_error,
    mean_squared_error,
    mean_squared_log_error,
    r2_score,
)

from krisi.evaluate.library.diagrams import (
    display_acf_plot,
    display_density_plot,
    display_single_value,
    display_time_series,
)
from krisi.evaluate.library.metric_wrappers import ljung_box
from krisi.evaluate.metric import Metric
from krisi.evaluate.type import ComputationalComplexity, MetricCategories, SampleTypes

mae = Metric[float](
    name="Mean Absolute Error",
    key="mae",
    category=MetricCategories.reg_err,
    info="(Mean absolute error) represents the difference between the original and predicted values extracted by averaged the absolute difference over the data set.",
    func=mean_absolute_error,
    plot_funcs=[display_single_value],
    plot_func_rolling=display_time_series,
)

mape = Metric[float](
    name="Mean Absolute Percentage Error",
    key="mape",
    category=MetricCategories.reg_err,
    func=mean_absolute_percentage_error,
    plot_funcs=[display_single_value],
    plot_func_rolling=display_time_series,
)

smape = Metric[float](
    name="Symmetric Mean Absolute Percentage Error",
    key="smape",
    category=MetricCategories.reg_err,
    func=lambda y_true, y_pred: np.mean(
        np.abs(y_pred - y_true) / (np.abs(y_true) + np.abs(y_pred)), axis=0
    ),
    plot_funcs=[display_single_value],
    plot_func_rolling=display_time_series,
)

mse = Metric[float](
    name="Mean Squared Error",
    key="mse",
    category=MetricCategories.reg_err,
    info="(Mean Squared Error) represents the difference between the original and predicted values extracted by squared the average difference over the data set.",
    parameters={"squared": True},
    func=mean_squared_error,
    plot_funcs=[display_single_value],
    plot_func_rolling=display_time_series,
)
rmse = Metric[float](
    name="Root Mean Squared Error",
    key="rmse",
    category=MetricCategories.reg_err,
    info="(Root Mean Squared Error) is the error rate by the square root of Mean Squared Error.",
    parameters={"squared": False},
    func=mean_squared_error,
    plot_funcs=[display_single_value],
    plot_func_rolling=display_time_series,
)
rmsle = Metric[float](
    name="Root Mean Squared Log Error",
    key="rmsle",
    category=MetricCategories.reg_err,
    parameters={"squared": False},
    func=mean_squared_log_error,
    plot_funcs=[display_single_value],
    plot_func_rolling=display_time_series,
)
r_two = Metric[float](
    name="R-squared",
    key="r_two",
    category=MetricCategories.reg_err,
    info="(Coefficient of determination) represents the coefficient of how well the values fit compared to the original values. The value from 0 to 1 interpreted as percentages. The higher the value is, the better the model is.",
    func=r2_score,
    plot_funcs=[display_single_value],
    plot_func_rolling=display_time_series,
)
residuals = Metric[List[float]](
    name="Residuals",
    key="residuals",
    category=MetricCategories.residual,
    func=lambda y, pred: y - pred,
    plot_func_rolling=display_time_series,
    plot_funcs=[display_acf_plot, display_density_plot, display_time_series],
)
residuals_mean = Metric[float](
    name="Mean of the Residuals",
    key="residuals_mean",
    category=MetricCategories.residual,
    func=lambda y, pred: (y - pred).mean(),
    plot_funcs=[display_single_value],
    plot_func_rolling=display_time_series,
)
residuals_std = Metric[float](
    name="Standard Deviation of the Residuals",
    key="residuals_std",
    category=MetricCategories.residual,
    func=lambda y, pred: (y - pred).std(),
    plot_funcs=[display_single_value],
    plot_func_rolling=display_time_series,
)
ljung_box_statistics = Metric[pd.DataFrame](
    name="Ljung Box Statistics",
    key="ljung_box_statistics",
    category=MetricCategories.residual,
    func=ljung_box,
    info="If p is larger than our significance level then we cannot dismiss the null-hypothesis that the residuals are a random walk.",
    restrict_to_sample=SampleTypes.insample,
    comp_complexity=ComputationalComplexity.high,
)

all_regression_metrics = [
    mae,
    mape,
    smape,
    mse,
    rmse,
    rmsle,
    r_two,
    residuals,
    residuals_mean,
    residuals_std,
    ljung_box_statistics,
]

minimal_regression_metrics = [
    mae,
    mape,
    mse,
    rmse,
    rmsle,
    r_two,
]

low_computation_regression_mterics = [
    metric
    for metric in all_regression_metrics
    if metric.comp_complexity is not ComputationalComplexity.high
]



from typing import Tuple

import pandas as pd
from sklearn.model_selection import train_test_split

from krisi.evaluate import ScoreCard, evaluate_in_outsample
from krisi.evaluate.type import Calculation
from krisi.utils.data import generating_arima_synthetic_data
from krisi.utils.iterable_helpers import type_converter
from krisi.utils.models import default_arima_model
from krisi.utils.runner import fit_model, generate_univariate_predictions


def basic_report_rolling() -> Tuple[ScoreCard, ScoreCard]:
    """Create syntethic Data"""
    target_col = "arima_synthetic"
    df = generating_arima_synthetic_data(target_col).to_frame(name=target_col)

    """ Split Data into Train and Test splits"""
    train, test = train_test_split(df, test_size=0.2, shuffle=False)
    train, test = type_converter(pd.DataFrame)([train, test])

    """ Train an ARIMA model """
    fit_model(default_arima_model, train)

    """ Generate Predictions with the model """
    insample_prediction, outsample_prediction = generate_univariate_predictions(
        default_arima_model, train, len(test), target_col
    )

    """ Evaluate predictions """
    report_insample, report_outsample = evaluate_in_outsample(
        model_name="default_arima_model",
        dataset_name="synthetic_arima",
        calculation=Calculation.rolling,
        y_insample=train[target_col],
        insample_predictions=insample_prediction,
        y_outsample=test[target_col],
        outsample_predictions=outsample_prediction,
    )

    """ Console log Reports """
    print(report_outsample)
    report_insample.print_summary()

    return report_insample, report_outsample


if __name__ == "__main__":
    basic_report_rolling()



from typing import TYPE_CHECKING, Any, Callable, Iterable, List, Union

import numpy as np
import pandas as pd

from krisi.evaluate.type import Predictions, Targets

if TYPE_CHECKING:
    from krisi.evaluate.metric import Metric
    from krisi.report.type import InteractiveFigure


def map_newdict_on_olddict(
    old_dict: dict, new_dict: dict, exclude: List[str] = []
) -> dict:
    merged_dict = old_dict.copy()
    for key_, value_ in new_dict.items():
        if key_ in old_dict and value_ is not None and key_ not in exclude:
            merged_dict[key_] = value_

    return merged_dict


def group_by_categories(
    flat_list: List[Union["Metric", "InteractiveFigure"]], categories: List[str]
) -> dict[str, Union["Metric", "InteractiveFigure"]]:
    category_groups = dict()
    for category in categories:
        category_groups[category] = list(
            filter(
                lambda x: x.category.value == category
                if hasattr(x, "category")
                else False,
                flat_list,
            )
        )
    category_groups[None] = list(
        filter(
            lambda x: x.category.value is None if hasattr(x, "category") else False,
            flat_list,
        )
    )
    return category_groups


def type_converter(type_to_ensure: Any) -> Callable:
    def ensure_format(
        obj_s: Union[List[Any], Any]
    ) -> Union[List[type_to_ensure], type_to_ensure]:
        if isinstance(obj_s, Iterable) and not isinstance(obj_s, str):
            return [
                type_to_ensure(el) if not isinstance(el, type_to_ensure) else el
                for el in obj_s
            ]
        else:
            if isinstance(obj_s, type_to_ensure):
                return obj_s
            else:
                return type_to_ensure(obj_s)

    return ensure_format


def string_to_id(s: str) -> str:
    s = "".join(filter(lambda c: str.isidentifier(c) or str.isdecimal(c), s))
    return s


def isiterable(obj: Any) -> bool:
    if isinstance(obj, Iterable) and not isinstance(obj, str):
        return True
    else:
        return False


def strip_builtin_functions(dict_to_strip: dict) -> dict:
    return {key_: value_ for key_, value_ in dict_to_strip.items() if key_[:2] != "__"}


def __flatten(xs: Iterable[Any]) -> Any:
    for x in xs:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x


def flatten(xs: Iterable[Any]) -> Iterable[Any]:
    return list(__flatten(xs))


def remove_nans(iter):
    if isinstance(iter, List):
        return [el for el in iter if el is not None]
    else:
        return {key: el for key, el in iter.items() if el is not None}


def calculate_nans(ds: Union[Targets, Predictions]) -> int:
    return ds.isna().sum() if isinstance(ds, pd.Series) else sum(np.isnan(ds))



import numpy as np

from krisi import score

score(
    y=np.random.randint(0, 2, 1000),
    predictions=np.random.randint(0, 2, 1000),
    # classification=True, # Optional, tries to decide based on if target contains integers
).print_summary()



from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd

from krisi.utils.models import Model


def fit_model(
    model: Model, train: pd.DataFrame, test: Optional[pd.Series] = None
) -> None:

    if test is None:
        test = pd.Series([])

    model.fit(train, test)


def generate_univariate_predictions(
    model: Model,
    train: pd.DataFrame,
    len_test: int,
    target_col: str,
) -> Tuple[Union[pd.Series, np.ndarray], Union[pd.Series, np.ndarray]]:

    insample_prediction = model.predict_in_sample(train[[target_col]])
    outsample_prediction = model.predict(pd.DataFrame([x for x in range(len_test)]))

    return insample_prediction, outsample_prediction



from dataclasses import dataclass, field
from enum import Enum
from typing import Callable, List, Literal, Optional, Union

import pandas as pd
import plotly.graph_objects as go
from dash import dcc

from krisi.evaluate.type import MetricCategories, MetricResult

PlotFunction = Callable[
    [List[MetricResult], str, Optional[float]],
    go.Figure,
]


@dataclass
class PlotlyInput:
    type: Literal[dcc.Dropdown, dcc.Input, dcc.Checklist, dcc.Slider, dcc.RangeSlider]
    id: str
    value_name: str
    default_value: Union[str, float, int]
    options: Optional[List[Union[str, float, int]]] = None


@dataclass
class InteractiveFigure:
    id: str
    get_figure: Callable
    inputs: List[PlotlyInput] = field(default_factory=list)
    global_input_ids: List[str] = field(default_factory=list)
    title: Optional[str] = ""
    width: Optional[float] = 900.0
    height: Optional[float] = 600.0
    category: Optional[MetricCategories] = None


class DisplayModes(Enum):
    interactive = "interactive"
    pdf = "pdf"
    direct = "direct"


class PathConst:
    default_save_path = "output/report/"
    html_template_url = "libary/default/template.html"
    css_template_url = "libary/default/template.css"


def plotly_interactive(
    plot_function: PlotFunction,
    data_source: Union[pd.DataFrame, pd.Series],
    *args,
    **kwargs
) -> Callable:
    # default_args = args
    default_kwargs = kwargs

    def wrapper(*args, **kwargs) -> go.Figure:
        width = kwargs.pop("width", None)
        height = kwargs.pop("height", None)
        title = kwargs.pop("title", None)

        for key, value in default_kwargs.items():
            if key not in kwargs:
                kwargs[key] = value

        fig = plot_function(data_source, *args, **kwargs)

        # fig.update_layout(width=width)
        # fig.update_layout(autosize=False, width=width, height=height)
        fig.update_layout(autosize=False, width=width, height=height)
        if title is not None:
            fig.update_layout(title=title)
        return fig

    return wrapper



import datetime
from typing import TYPE_CHECKING, Callable, List, Optional, Tuple, Union

from krisi.evaluate.type import MetricCategories, ScoreCardMetadata
from krisi.report.interactive import run_app
from krisi.report.pdf import append_sizes, convert_figures_to_html, create_pdf_report
from krisi.report.type import DisplayModes, InteractiveFigure, PlotlyInput
from krisi.utils.iterable_helpers import flatten, group_by_categories, remove_nans

if TYPE_CHECKING:
    from krisi.evaluate.metric import Metric
    from krisi.evaluate.scorecard import ScoreCard


class Report:
    title: str
    modes: List[Union[str, DisplayModes]]

    def __init__(
        self,
        title: str,
        general_description: str = "",
        modes: Union[str, List[str], DisplayModes, List[DisplayModes]] = [
            DisplayModes.pdf
        ],
        figures: List[InteractiveFigure] = [],
        global_controllers: List[PlotlyInput] = [],
        html_template_url: str = "library/default/template.html",
        css_template_url: str = "library/default/template.css",
        get_html_elements: Optional[Callable] = None,
        scorecard_metadata: Optional[ScoreCardMetadata] = None,
    ) -> None:
        self.title = title
        self.general_description = general_description
        self.scorecard_metadata = scorecard_metadata
        if isinstance(modes, (str, DisplayModes)):
            self.modes = [modes]
        else:
            self.modes = modes
        self.figures = figures
        self.global_controllers = global_controllers
        self.html_template_url = html_template_url
        self.css_template_url = css_template_url
        self.get_html_elements = get_html_elements

    def generate_launch(self) -> None:
        figures_by_category = group_by_categories(
            self.figures, [el.value for el in MetricCategories]
        )

        if DisplayModes.pdf in self.modes or DisplayModes.pdf.value in self.modes:
            create_pdf_report(
                html_template_url=self.html_template_url,
                css_template_url=self.css_template_url,
                html_elements_to_inject=self.get_html_elements()
                if self.get_html_elements is not None
                else dict(),
            )

        if DisplayModes.direct in self.modes or DisplayModes.direct.value in self.modes:
            [figure.get_figure(width=900.0).show() for figure in self.figures]

        if (
            DisplayModes.interactive in self.modes
            or DisplayModes.interactive.value in self.modes
        ):
            run_app(
                figures_by_category,
                self.global_controllers,
                self.title,
                self.general_description,
                self.scorecard_metadata,
                # stop_event=stop_event,
            )


def get_all_interactive_diagrams(metrics: List["Metric"]) -> List[InteractiveFigure]:
    interactive_diagrams = remove_nans(
        [metric.get_diagrams() for metric in metrics]
        + [metric.get_diagram_over_time() for metric in metrics]
    )
    return flatten(interactive_diagrams)


def get_waterfall_metric_html(
    metrics: List["Metric"],
) -> Tuple[str, List[InteractiveFigure]]:
    interactive_diagrams = get_all_interactive_diagrams(metrics)

    html_images = convert_figures_to_html(
        [
            interactive_diagram.get_figure()
            for interactive_diagram in interactive_diagrams
        ]
    )

    return html_images, interactive_diagrams


def get_html_elements_for_injection_scorecard(
    obj: "ScoreCard",
    author: str,
    project_name: str,
    date: str,
    custom_metric_html: str,
) -> Callable:
    def func() -> dict[str, str]:

        diagrams = obj.get_diagram_dictionary()
        diagrams = append_sizes(diagrams)

        diagrams_static = {
            interactive_figure.id: convert_figures_to_html(
                [
                    interactive_figure.get_figure(
                        width=interactive_figure.width,
                        height=interactive_figure.height,
                        title=interactive_figure.title,
                    )
                ]
            )
            for key, interactive_figure in diagrams.items()
        }

        return dict(
            author=author,
            project_name=project_name,
            date=date,
            custom_metric_html=custom_metric_html,
            **diagrams_static,
        )

    return func


def create_report_from_scorecard(
    obj: "ScoreCard",
    display_modes: Union[str, List[str], DisplayModes, List[DisplayModes]],
    html_template_url: str,
    css_template_url: str,
    author: str,
) -> Report:
    if isinstance(display_modes, (str, DisplayModes)):
        display_modes_ = [display_modes]
    else:
        display_modes_ = display_modes

    custom_metric_html, interactive_figures = get_waterfall_metric_html(
        obj.get_all_metrics()
    )

    get_html_elements = None
    if DisplayModes.pdf in display_modes_ or DisplayModes.pdf.value in display_modes_:
        get_html_elements = get_html_elements_for_injection_scorecard(
            obj=obj,
            author=author,
            project_name=obj.metadata.project_name,
            date=datetime.datetime.now().strftime("%Y-%m-%d"),
            custom_metric_html=custom_metric_html,
        )

    return Report(
        title=f"Report on {obj.metadata.model_name}",
        general_description="General Description",
        modes=display_modes_,
        scorecard_metadata=obj.metadata,
        figures=interactive_figures,
        html_template_url=html_template_url,
        css_template_url=css_template_url,
        get_html_elements=get_html_elements,
    )



import numpy as np

from krisi import compare, score

scorecards = [
    score(y=np.random.rand(1000), predictions=np.random.rand(1000)) for _ in range(5)
]
compare(scorecards, sort_metric_key="rmse", metrics_to_display=["mse", "mae"])



from typing import List

import numpy as np
import pandas as pd
from statsmodels.tsa.arima_process import ArmaProcess


def make_it_stationary(ds: pd.Series) -> pd.Series:
    """If the ds, is a random walk with drift, take first differences to make it stationary."""
    print(f"Making the Series {ds.name} stationary.")
    return ds.diff().dropna()


def generating_arima_synthetic_data(
    target_col: str,
    nsample: int = 1000,
    ar: List[float] = [1.0, -0.9],
    ma: List[float] = [1.0],
) -> pd.Series:
    """Generate Synthetic data with ARIMA process."""
    print(f"Generating Synthetic ARIMA process with nsamples:{nsample}")

    # Plot 1: AR parameter = +0.9,
    # exponent of PHI is opposite due to original signal processing literature convention
    ar = np.array(ar)
    ma = np.array(ma)
    AR_object = ArmaProcess(ar, ma)
    simulated_data = AR_object.generate_sample(nsample=nsample)

    return pd.Series(
        simulated_data,
        name=target_col,
        index=pd.date_range(end="2022", periods=nsample, freq="D"),
    )



import os

import numpy as np

from krisi import ScoreCard
from krisi.evaluate.type import PathConst
from krisi.utils.io import load_scorecards


def create_scorecard(project_name: str) -> None:
    target, predictions = np.random.rand(100), np.random.rand(100)

    sc = ScoreCard(
        target,
        predictions,
        model_name="<your_model_name>",
        dataset_name="<your_dataset_name>",
        project_name=project_name,
    )

    sc.evaluate(defaults=True)
    sc.save()


def load_and_generate(project_name: str) -> None:
    scorecard = load_scorecards(project_name, PathConst.default_eval_output_path)
    scorecard[-1].generate_report()


project_name = "Example Project"
create_scorecard(project_name)


if "PYTEST_CURRENT_TEST" in os.environ:
    print("Not testing Dash server currently")
else:
    load_and_generate(project_name)



from dataclasses import dataclass
from enum import Enum
from typing import Callable, List, Tuple, TypeVar, Union

import numpy as np
import pandas as pd

MetricResult = TypeVar(
    "MetricResult", bound=Union[float, int, str, List, Tuple, pd.DataFrame]
)


Predictions = Union[np.ndarray, pd.Series, List[Union[int, float]]]
Targets = Union[np.ndarray, pd.Series, List[Union[int, float]]]
MetricFunction = Callable[[Predictions, Targets], MetricResult]


class MetricCategories(Enum):
    residual = "Residual Diagnostics"
    entropy = "Information Entropy"
    class_err = "Forecast Errors - Classification"
    reg_err = "Forecast Errors - Regression"
    unknown = "Unknown"


class SampleTypes(Enum):
    insample = "insample"
    outofsample = "outofsample"


class Calculation(Enum):
    single = "single"
    rolling = "rolling"
    both = "both"


class SaveModes(Enum):
    svg = "svg"
    html = "html"
    text = "text"
    obj = "obj"
    minimal = "minimal"


class PathConst:
    default_eval_output_path: str = "output/"
    html_report_template_url: str = "library/pdf_layouts/scorecard/report.html"
    css_report_template_url: str = "library/pdf_layouts/scorecard/report.css"


class ComputationalComplexity(Enum):
    low = "low"
    medium = "medium"
    high = "high"


@dataclass
class ScoreCardMetadata:
    project_name: str = ""
    project_description: str = ""
    model_name: str = ""
    model_description: str = ""
    dataset_name: str = ""
    dataset_description: str = ""



import numpy as np

from krisi.evaluate import ScoreCard
from krisi.report.type import DisplayModes

target, predictions = np.random.rand(100), np.random.rand(100)
sc = ScoreCard(
    target,
    predictions,
    model_name="<your_model_name>",
    dataset_name="<your_dataset_name>",
    project_name="Example Project",
)

""" A predefined metrics """
sc.evaluate(defaults=True)
sc.generate_report(
    display_modes=[DisplayModes.pdf],
)



import numpy as np

from krisi import compare, score
from krisi.evaluate.type import PathConst
from krisi.utils.io import load_scorecards

project_name = "Example Load"

for _ in range(5):
    score(
        y=np.random.rand(1000),
        predictions=np.random.rand(1000),
        project_name=project_name,
    ).save()


compare(
    load_scorecards(project_name, PathConst.default_eval_output_path),
    metrics_to_display=["mse"],
)



import numpy as np

from krisi.evaluate import Metric, MetricCategories, SampleTypes, ScoreCard

target, predictions = np.random.rand(100), np.random.rand(100)

sc = ScoreCard(
    target,
    predictions,
    "<your_model_name>",
    dataset_name="<your_dataset_name>",
    sample_type=SampleTypes.insample,
)

""" Predefined metrics """
sc.evaluate(defaults=True)

""" Adding a new metric """
calculated_metric_example = (target - predictions).mean()

sc["own_metric"] = calculated_metric_example
# or
sc.another_metric = calculated_metric_example * 2.0
# or
sc["yet_another_metric"] = Metric(
    name="A new, own Metric",
    category=MetricCategories.residual,
    result=calculated_metric_example * 3.0,
    parameters={"hyper_1": 5.0},
)

# Updating a metric
sc.yet_another_metric = dict(info="Giving description to a metric")

""" Print scorecard summary """
sc.print_summary(with_info=True, extended=True)



import datetime
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from rich import print
from rich.pretty import Pretty

from krisi.evaluate.assertions import is_dataset_classification_like
from krisi.evaluate.library.default_metrics_classification import (
    all_classification_metrics,
)
from krisi.evaluate.library.default_metrics_regression import all_regression_metrics
from krisi.evaluate.metric import Metric
from krisi.evaluate.type import (
    MetricCategories,
    PathConst,
    Predictions,
    SampleTypes,
    SaveModes,
    ScoreCardMetadata,
    Targets,
)
from krisi.evaluate.utils import handle_unnamed
from krisi.report.console import get_minimal_summary, get_summary
from krisi.report.report import create_report_from_scorecard
from krisi.report.type import DisplayModes, InteractiveFigure
from krisi.utils.io import save_console, save_minimal_summary, save_object
from krisi.utils.iterable_helpers import (
    flatten,
    map_newdict_on_olddict,
    remove_nans,
    strip_builtin_functions,
)


@dataclass
class ScoreCard:
    """ScoreCard Object.

    Krisi's main object holding and evaluating metrics.
    Stores, evaluates and generates vizualisations of predefined
    and custom metrics for regression and classification.

    Examples
    --------
    >>> from krisi import ScoreCard
    ... y_pred, y_true = [0, 2, 1, 3], [0, 1, 2, 3]
    ... sc = ScoreCard()
    ... sc.evaluate(y_pred, y_true, defaults=True) # Calculate predefined metrics
    ... sc["own_metric"] = (y_pred - y_true).mean() # Add a metric result directly
    ... sc.print_summary(extended=True)
    """

    y: Targets
    predictions: Predictions
    sample_type: SampleTypes
    default_metrics_keys: List[str]
    custom_metrics_keys: List[str]
    classification: bool  # TODO: Support multilabel classification
    metadata: ScoreCardMetadata

    def __init__(
        self,
        y: Targets,
        predictions: Predictions,
        model_name: Optional[str] = None,
        model_description: str = "",
        dataset_name: Optional[str] = None,
        dataset_description: str = "",
        project_name: Optional[str] = None,
        project_description: str = "",
        classification: Optional[bool] = None,
        sample_type: SampleTypes = SampleTypes.outofsample,
        default_metrics: Optional[List[Metric]] = None,
        custom_metrics: Optional[List[Metric]] = None,
    ) -> None:
        self.__dict__["y"] = y
        self.__dict__["predictions"] = predictions
        self.__dict__["sample_type"] = sample_type
        self.__dict__["classification"] = (
            is_dataset_classification_like(y)
            if classification is None
            else classification
        )
        model_name_, dataset_name_, project_name_ = handle_unnamed(
            y, model_name, dataset_name, project_name
        )

        self.__dict__["metadata"] = ScoreCardMetadata(
            project_name_,
            project_description,
            model_name_,
            model_description,
            dataset_name_,
            dataset_description,
        )

        if default_metrics is None:
            default_metrics = (
                all_classification_metrics
                if self.classification
                else all_regression_metrics
            )

        self.__dict__["default_metrics_keys"] = [
            metric.key for metric in default_metrics
        ]
        if custom_metrics is None:
            custom_metrics = []

        self.__dict__["custom_metrics_keys"] = [metric.key for metric in custom_metrics]

        for metric in default_metrics:
            self.__dict__[metric.key] = deepcopy(metric)

        for metric in custom_metrics:
            self.__dict__[metric.key] = deepcopy(metric)

    def __setitem__(self, key: str, item: Any) -> None:
        self.__setattr__(key, item)

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key, "Unknown metric")

    def __delitem__(self, key: str) -> None:
        setattr(self, key, None)

    def __str__(self) -> str:
        print(Pretty(self.__dict__))
        return ""

    def __repr__(self) -> str:
        print(Pretty(self.__dict__))
        return ""

    def __setattr__(self, key: str, item: Any) -> None:
        """Defines Dictionary like behaviour and ensures that a Metric can be
        added as a
            - Metric object,
            - Dictionary,
            - Direct result (float, int or a List of float or int). Gets wrapped in a ``Metric`` object

        See examples for behaviour.

        Parameters
        ----------
        key : string
            The key to which the object will be assign to on this object

        item : Dictionary, Metric, Float, Int or List of Float or Int, or pd.Series
            Always ignored, exists for compatibility.

        Returns
        -------
        None

        Examples
        --------
        >>> from krisi import ScoreCard
        ... sc = ScoreCard()
        ... sc['metric_result'] = 0.53 # Direct result assignment as a Dictionary
        Metric(result=0.53, key='metric_result', category=None, parameters=None, info="", ...)

        >>> sc.another_metric_result = 1 # Direct object assignment
        Metric(result=1, key='another_metric_result', category=None, parameters=None, info="", ...)

        >>> from krisi.evaluate.metric import Metric
        ... from krisi.evaluate.type import MetricCategories
        ... sc.full_metric = Metric("My own metric", category=MetricCategories.class_err, info="A fictious metric with metadata")
        Metric("My own metric", key="my_own_metric", func: lambda y, y_hat: (y - y_hat)/2, category=MetricCategories.class_err, info="A fictious metric with metadata", ...)

        >>> sc.metric_as_dictionary = {name: "My other metric", info: "A Metric created with a dictionary", func: lambda y, y_hat: y - y_hat}
        Metric("My other metric", key="my_other_metric", info="A Metric created with a dictionary", ...)

        """
        metric = getattr(self, key, None)

        if metric is None:
            if isinstance(item, dict):
                self.__dict__[key] = Metric(key=key, **item)
            elif isinstance(item, Metric):
                item["key"] = key
                self.__dict__[key] = item
            else:
                self.__dict__[key] = Metric(
                    name=key, key=key, result=item, category=MetricCategories.unknown
                )
        else:
            if isinstance(item, dict):
                metric_dict = map_newdict_on_olddict(
                    strip_builtin_functions(vars(metric)), item, exclude=["name"]
                )
                self.__dict__[key] = Metric(**metric_dict)
            elif isinstance(item, Metric):
                metric_dict = map_newdict_on_olddict(
                    strip_builtin_functions(vars(metric)),
                    strip_builtin_functions(vars(item)),
                    exclude=["name"],
                )
                self.__dict__[key] = Metric(**metric_dict)
            else:
                metric["result"] = item
                self.__dict__[key] = metric

    def get_default_metrics(self) -> List[Metric]:
        """Returns a List of Predefined Metrics according to task type:
        regression, classification, multi-label classification.


        Returns
        -------
        List of Metrics
        """
        return [self.__dict__[key] for key in self.default_metrics_keys]

    def get_custom_metrics(self) -> List[Metric]:
        """Returns a List of Custom ``Metric``s defined by the user on initalization
        of the ``ScoreCard``

        Returns
        -------
        List of Metrics
        """
        predifined_custom_metrics = [
            self.__dict__[key] for key in self.custom_metrics_keys
        ]
        modified_custom_metrics = [
            value
            for key, value in self.__dict__.items()
            if key not in self.custom_metrics_keys + self.default_metrics_keys
            and hasattr(value, "key")
        ]

        return predifined_custom_metrics + modified_custom_metrics

    def get_all_metrics(self, defaults: bool = True) -> List[Metric]:
        if defaults:
            return self.get_default_metrics() + self.get_custom_metrics()
        else:
            return self.get_custom_metrics()

    def evaluate(self, defaults: bool = True) -> "ScoreCard":
        """Evaluates ``Metric``s present on the ``ScoreCard``

        Parameters
        ----------
        y: Targets = Union[np.ndarray, pd.Series, List[Union[int, float]]]
        The true labels to compare values to

        predictions: Predictions = Union[np.ndarray, pd.Series, List[Union[int, float]]]
        The predicted values. Integers or whole floats if classification, else floats.

        defaults: boolean
        Wether the default ``Metric``s should be evaluated or not. Default value = True

        Returns
        -------
        self: ScoreCard
        """

        for metric in self.get_all_metrics(defaults=defaults):
            if metric.restrict_to_sample is not self.sample_type:
                metric.evaluate(self.y, self.predictions)

        return self

    def evaluate_over_time(
        self,
        defaults: bool = True,
        window: Optional[int] = None,
    ) -> "ScoreCard":
        """Evaluates ``Metric``s present on the ``ScoreCard`` over time, either with expanding
        or fixed sized window. Assigns list of results to ``results_over_time``.

        Parameters
        ----------
        y: Targets = Union[np.ndarray, pd.Series, List[Union[int, float]]]
        The true labels to compare values to

        predictions: Predictions = Union[np.ndarray, pd.Series, List[Union[int, float]]]
        The predicted values. Integers or whole floats if classification, else floats.

        defaults: boolean
        Wether the default ``Metric``s should be evaluated or not. Default value = True

        window: int - Optional
        Size of window. If number is provided then evaluation happens on a fixed window size,
        otherwise it evaluates it on an expanding window basis.

        Returns
        -------
        self: ScoreCard
        """
        for metric in self.get_all_metrics(defaults=defaults):
            if metric.restrict_to_sample is not self.sample_type:
                metric.evaluate_over_time(self.y, self.predictions, window=window)
        return self

    def print_summary(
        self,
        with_info: bool = False,
        extended: bool = True,
        input_analysis: bool = True,
    ) -> None:
        if extended:
            summary = get_summary(
                self,
                repr=True,
                categories=[el.value for el in MetricCategories],
                with_info=with_info,
                input_analysis=input_analysis,
            )
        else:
            summary = get_minimal_summary(self)

        print(summary)

    def save(
        self,
        path: str = PathConst.default_eval_output_path,
        with_info: bool = False,
        save_modes: List[Union[SaveModes, str]] = [
            SaveModes.minimal,
            SaveModes.obj,
            SaveModes.text,
        ],
    ) -> "ScoreCard":
        if self.metadata.project_name:
            path += f"{self.metadata.project_name}/"
        path += f"{datetime.datetime.now().strftime('%H:%M:%S')}_{self.metadata.model_name}_{self.metadata.dataset_name}"
        import os

        if not os.path.exists(path):
            os.makedirs(path)

        if SaveModes.minimal in save_modes or SaveModes.minimal.value in save_modes:
            save_minimal_summary(self, path)
        if SaveModes.obj in save_modes or SaveModes.obj.value in save_modes:
            save_object(self, path)
        save_console(self, path, with_info, save_modes)

        return self

    def get_diagram_dictionary(self) -> Dict[str, InteractiveFigure]:
        diagrams = flatten(
            remove_nans([metric.get_diagrams() for metric in self.get_all_metrics()])
        )

        return {diagram.id: diagram for diagram in diagrams}

    def generate_report(
        self,
        display_modes: Union[str, List[str], DisplayModes, List[DisplayModes]] = [
            DisplayModes.interactive
        ],
        html_template_url: str = PathConst.html_report_template_url,
        css_template_url: str = PathConst.css_report_template_url,
        author: str = "",
    ) -> None:
        report = create_report_from_scorecard(
            self,
            display_modes,
            html_template_url,
            css_template_url,
            author=author,
        )
        report.generate_launch()


def get_rolling_diagrams(obj: "ScoreCard") -> List[List[InteractiveFigure]]:
    return [
        diagram
        for diagram in [
            metric.get_diagram_over_time() for metric in obj.get_all_metrics()
        ]
        if diagram is not None
    ]



from typing import Any, Tuple, Union

import pandas as pd
from statsmodels.stats.diagnostic import acorr_ljungbox

from krisi.evaluate.type import Predictions, Targets


def ljung_box(
    y: Targets, predictions: Predictions
) -> Union[pd.DataFrame, Tuple[Any, Any], Tuple[Any, Any, Any, Any]]:
    return acorr_ljungbox(predictions)



from typing import List, Optional, Tuple, Union

import pandas as pd

from krisi.evaluate.metric import Metric
from krisi.evaluate.scorecard import ScoreCard
from krisi.evaluate.type import Calculation, Predictions, SampleTypes, Targets


def score(
    y: Targets,
    predictions: Predictions,
    model_name: Optional[str] = None,
    dataset_name: Optional[str] = None,
    project_name: Optional[str] = None,
    default_metrics: Optional[List[Metric]] = None,
    custom_metrics: Optional[List[Metric]] = None,
    classification: Optional[bool] = None,
    sample_type: SampleTypes = SampleTypes.outofsample,
    calculation: Union[Calculation, str] = Calculation.single,
    window: int = 30,
) -> ScoreCard:
    """_summary_

    Args:
        y (Targets): _description_
        predictions (Predictions): _description_
        model_name (Optional[str], optional): _description_. Defaults to None.
        dataset_name (Optional[str], optional): _description_. Defaults to None.
        project_name (Optional[str], optional): _description_. Defaults to None.
        default_metrics (Optional[List[Metric]], optional): _description_. Defaults to None.
        custom_metrics (Optional[List[Metric]], optional): _description_. Defaults to None.
        classification (Optional[bool], optional): _description_. Defaults to None.
        sample_type (SampleTypes, optional): _description_. Defaults to SampleTypes.outofsample.
        calculation (Union[Calculation, str], optional): _description_. Defaults to Calculation.single.
        window (int, optional): _description_. Defaults to 30.

    Raises:
        ValueError: _description_

    Returns:
        ScoreCard: _description_
    """

    sc = ScoreCard(
        y=y,
        predictions=predictions,
        model_name=model_name,
        dataset_name=dataset_name,
        project_name=project_name,
        sample_type=sample_type,
        classification=classification,
        default_metrics=default_metrics,
        custom_metrics=custom_metrics,
    )

    if calculation == Calculation.single or calculation == Calculation.single.value:
        sc.evaluate()
    elif calculation == Calculation.rolling or calculation == Calculation.rolling.value:
        sc.evaluate_over_time(window=window)
    elif calculation == Calculation.both or calculation == Calculation.both.value:
        sc.evaluate()
        sc.evaluate_over_time(window=window)
    else:
        raise ValueError(f"Calculation type {calculation} not recognized.")

    return sc


def evaluate_in_outsample(
    y_insample: pd.Series,
    insample_predictions: pd.Series,
    y_outsample: pd.Series,
    outsample_predictions: pd.Series,
    model_name: str = "Unknown model",
    dataset_name: Optional[str] = None,
    project_name: Optional[str] = None,
    default_metrics: Optional[List[Metric]] = None,
    custom_metrics: Optional[List[Metric]] = None,
    classification: Optional[bool] = None,
    calculation: Union[Calculation, str] = Calculation.single,
) -> Tuple[ScoreCard, ScoreCard]:

    insample_summary = score(
        y_insample,
        insample_predictions,
        model_name,
        dataset_name,
        project_name,
        default_metrics,
        custom_metrics,
        classification,
        SampleTypes.insample,
        calculation,
    )
    outsample_summary = score(
        y_outsample,
        outsample_predictions,
        model_name,
        dataset_name,
        project_name,
        default_metrics,
        custom_metrics,
        classification,
        SampleTypes.outofsample,
        calculation,
    )

    return insample_summary, outsample_summary



from typing import List

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from statsmodels.tsa.stattools import acf, pacf

from krisi.evaluate.type import MetricResult


def display_time_series(data: List[MetricResult], title: str = "") -> go.Figure:
    df = pd.DataFrame(data, columns=[title])
    df["iteration"] = list(range(len(data)))
    fig = px.line(
        df,
        x="iteration",
        y=title,
    )
    fig.update_layout(title=title)
    return fig


def display_single_value(data: MetricResult, title: str = "") -> go.Figure:

    fig = go.Figure()

    fig.add_trace(
        go.Indicator(
            value=data,
            delta={},
            gauge={"axis": {"visible": False}},
            domain={"row": 0, "column": 0},
        )
    )

    fig.update_layout(title=title)
    return fig


def display_acf_plot(
    data: MetricResult,
    title: str = "",
    plot_pacf: bool = False,
) -> go.Figure:

    title = (
        title + " - Partial Autocorrelation (PACF)"
        if plot_pacf
        else "Autocorrelation (ACF)"
    )

    if not isinstance(data, pd.Series):
        data = pd.Series(data)

    corr_array = (
        pacf(data.dropna(), alpha=0.05) if plot_pacf else acf(data.dropna(), alpha=0.05)
    )
    lower_y = corr_array[1][:, 0] - corr_array[0]
    upper_y = corr_array[1][:, 1] - corr_array[0]

    fig = go.Figure()
    [
        fig.add_scatter(
            x=(x, x), y=(0, corr_array[0][x]), mode="lines", line_color="#3f3f3f"
        )
        for x in range(len(corr_array[0]))
    ]
    fig.add_scatter(
        x=np.arange(len(corr_array[0])),
        y=corr_array[0],
        mode="markers",
        marker_color="#1f77b4",
        marker_size=12,
    )
    fig.add_scatter(
        x=np.arange(len(corr_array[0])),
        y=upper_y,
        mode="lines",
        line_color="rgba(255,255,255,0)",
    )
    fig.add_scatter(
        x=np.arange(len(corr_array[0])),
        y=lower_y,
        mode="lines",
        fillcolor="rgba(32, 146, 230,0.3)",
        fill="tonexty",
        line_color="rgba(255,255,255,0)",
    )
    fig.update_traces(showlegend=False)
    fig.update_xaxes(range=[-1, 42])
    fig.update_yaxes(zerolinecolor="#000000")

    fig.update_layout(title=title)
    return fig


def display_density_plot(
    data: MetricResult,
    title: str = "",
    plot_pacf: bool = False,
) -> go.Figure:
    if not isinstance(data, pd.Series):
        data = pd.Series(data)
    fig = px.histogram(data, marginal="box")  # or violin, rug
    fig.update_layout(title=title)

    return fig



import logging
from dataclasses import dataclass, field
from typing import Any, Generic, List, Optional, Union

from krisi.evaluate.assertions import check_valid_pred_target
from krisi.evaluate.type import (
    ComputationalComplexity,
    MetricCategories,
    MetricFunction,
    MetricResult,
    Predictions,
    SampleTypes,
    Targets,
)
from krisi.report.console import print_metric
from krisi.report.type import InteractiveFigure, PlotFunction, plotly_interactive
from krisi.utils.iterable_helpers import isiterable, string_to_id


@dataclass
class Metric(Generic[MetricResult]):
    name: str
    key: str = ""
    category: Optional[MetricCategories] = None
    result: Optional[Union[Exception, MetricResult, List[MetricResult]]] = None
    result_rolling: Optional[Union[Exception, MetricResult, List[MetricResult]]] = None
    parameters: dict = field(default_factory=dict)
    func: MetricFunction = lambda x, y: None
    plot_funcs: Optional[List[PlotFunction]] = None
    plot_func_rolling: Optional[PlotFunction] = None
    info: str = ""
    restrict_to_sample: Optional[SampleTypes] = None
    comp_complexity: Optional[ComputationalComplexity] = None

    def __post_init__(self):
        if self.key == "":
            self.key = string_to_id(self.name)

    def __setitem__(self, key: str, item: Any) -> None:
        setattr(self, key, item)

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key, "Unknown Field")

    def __str__(self) -> str:
        return print_metric(self)

    def __repr__(self) -> str:
        return print_metric(self, repr=True)

    def evaluate(self, y: Targets, predictions: Predictions) -> None:
        check_valid_pred_target(y, predictions)

        try:
            result = self.func(y, predictions, **self.parameters)
        except Exception as e:
            result = e
        self.__safe_set(result, key="result")

    def evaluate_over_time(
        self, y: Targets, predictions: Predictions, window: Optional[int] = None
    ) -> None:
        try:
            if window:
                result_rolling = [
                    self.func(
                        y[i : i + window],
                        predictions[i : i + window],
                        **self.parameters,
                    )
                    for i in range(0, len(y) - 1, window)
                ]
            else:
                # expanding
                result_rolling = [
                    self.func(y[: i + 1], predictions[: i + 1], **self.parameters)
                    for i in range(len(y) - 1)
                ]
        except Exception as e:
            result_rolling = e

        self.__safe_set(result_rolling, key="result_rolling")

    def get_diagram_over_time(self) -> Optional[List[InteractiveFigure]]:
        return create_diagram_rolling(self)

    def get_diagrams(
        self,
    ) -> Optional[List[InteractiveFigure]]:
        return create_diagram(self)

    def __safe_set(
        self, result: Union[Exception, MetricResult, List[MetricResult]], key: str
    ):
        if self.__dict__[key] is not None:
            raise ValueError("This metric already contains a result.")
        else:
            self.__dict__[key] = result


def create_diagram_rolling(obj: Metric) -> Optional[List[InteractiveFigure]]:
    if obj.plot_func_rolling is None:
        logging.info("No plot_func_rolling (Plotting Function Rolling) specified")
        return None
    elif isinstance(obj.result_rolling, Exception) or obj.result_rolling is None:
        return None
    elif isiterable(obj.result_rolling):
        return [
            InteractiveFigure(
                f"{obj.key}_{obj.plot_func_rolling.__name__}",
                get_figure=plotly_interactive(
                    obj.plot_func_rolling, obj.result_rolling, title=obj.name
                ),
                category=obj.category,
            )
        ]
    else:
        return None


def create_diagram(obj: Metric) -> Optional[List[InteractiveFigure]]:
    if obj.plot_funcs is None:
        logging.info("No plot_func (Plotting Function) specified")
        return None
    elif isinstance(obj.result, Exception) or obj.result is None:
        return None
    else:
        return [
            InteractiveFigure(
                f"{obj.key}_{plot_func.__name__}",
                get_figure=plotly_interactive(plot_func, obj.result, title=obj.name),
                title=f"{obj.name} - {plot_func.__name__}",
                category=obj.category,
            )
            for plot_func in obj.plot_funcs
        ]



from typing import TYPE_CHECKING, List, Union

from rich.console import Console

from krisi.evaluate.type import MetricCategories, PathConst, SaveModes
from krisi.report.console import get_minimal_summary, get_summary

if TYPE_CHECKING:
    from krisi.evaluate.scorecard import ScoreCard


def save_object(obj: "ScoreCard", path: str) -> None:
    import dill

    with open(f"{path}/scorecard.pickle", "wb") as file:
        dill.dump(obj, file)


def save_console(
    obj: "ScoreCard",
    path: str,
    with_info: bool,
    save_modes: List[Union[SaveModes, str]],
) -> None:

    summary = get_summary(
        obj,
        repr=True,
        categories=[el.value for el in MetricCategories],
        with_info=with_info,
    )

    console = Console(record=True, width=120)
    with console.capture():
        console.print(summary)

    if SaveModes.text in save_modes or SaveModes.text.value in save_modes:
        console.save_text(f"{path}/console.txt", clear=False)
    if SaveModes.html in save_modes or SaveModes.html.value in save_modes:
        console.save_html(f"{path}/console.html", clear=False)
    if SaveModes.svg in save_modes or SaveModes.svg.value in save_modes:
        console.save_svg(f"{path}/console.svg", title="save_table_svg.py", clear=False)


def save_minimal_summary(obj: "ScoreCard", path: str) -> None:
    text_summary = get_minimal_summary(obj)

    with open(f"{path}/minimal.txt", "w") as f:
        f.write(text_summary)


def load_scorecards(
    project_name: str, path: str = PathConst.default_eval_output_path
) -> List["ScoreCard"]:
    import os
    import pickle

    files = os.listdir(f"{path}/{project_name}")

    loaded_scorecards = []
    for file in files:
        with open(f"{path}{project_name}/{file}/scorecard.pickle", "rb") as f:
            loaded_scorecards.append(pickle.load(f))

    return loaded_scorecards



import os
from copy import deepcopy
from typing import TYPE_CHECKING, List, Union

import numpy as np
import pandas as pd
from rich import box
from rich.pretty import Pretty
from rich.table import Table

from krisi.evaluate.type import Predictions, Targets
from krisi.report.library.console.diagrams import (
    distribution_plot,
    histogram_plot,
    line_plot,
)
from krisi.utils.console_plot import plotextMixin
from krisi.utils.iterable_helpers import calculate_nans, isiterable

if TYPE_CHECKING:
    from krisi.evaluate.metric import Metric


def bold(text: str, rich: bool = True) -> str:
    return f"[bold]{text}[/bold]" if rich else f"\033[1m{text}\033[0m"


def get_term_size() -> int:
    term_size = os.get_terminal_size()
    return term_size.columns


def __display_result(metric: "Metric") -> Union[Pretty, plotextMixin]:
    if metric.result is None:
        result = deepcopy(metric.result_rolling)
    else:
        result = deepcopy(metric.result)

    if isinstance(result, Exception):
        result = str(result)
    elif isinstance(result, float):
        result = round(result, 3)

    if isiterable(result):
        if isinstance(result, (pd.Series, pd.DataFrame)):
            return Pretty(result)
        elif isiterable(result[0]):
            return Pretty("Result is a complex Iterable")
        else:
            # Create a Console Plot
            return plotextMixin(result, line_plot, title=metric.name)
    else:
        return Pretty(result, max_depth=2, max_length=3)


def __create_metric(metric: "Metric", with_info: bool) -> List[str]:
    metric_summarized = [
        f"{metric.name} ({metric.key})",
        __display_result(metric),
        Pretty(metric.parameters),
        Pretty(metric.info),
    ]
    metric_summarized = metric_summarized if with_info else metric_summarized[:-1]

    return metric_summarized


def create_metric_table(
    title: str, metrics: List["Metric"], with_info: bool, show_header: bool = False
) -> Table:
    table = Table(
        title=title,
        # show_edge=False,
        show_footer=False,
        show_lines=True,
        show_header=show_header,
        expand=True,
        box=box.ROUNDED,
    )

    table.add_column(
        "Metric Name", justify="right", style="cyan", width=1, no_wrap=False
    )
    table.add_column("Result", style="magenta", width=5)
    table.add_column("Parameters", style="green", width=1)
    if with_info:
        table.add_column("Info", width=3)

    for metric in metrics:
        if metric.result is None and metric.result_rolling is None:
            continue
        metric_summarized = __create_metric(metric, with_info)
        table.add_row(*metric_summarized)

    return table


def metrics_empty_in_category(metrics: List["Metric"]) -> bool:
    return (
        metrics is None
        or len(metrics) < 1
        or (
            all([metric.result is None for metric in metrics])
            and all([metric.result_rolling is None for metric in metrics])
        )
    )


def create_y_pred_table(classification: bool, y: Targets, preds: Predictions) -> Table:
    if isinstance(y, np.ndarray):
        y = pd.Series(y)
    if isinstance(preds, np.ndarray):
        preds = pd.Series(preds)

    table = Table(
        title="Targets and Predictions Analysis",
        # show_edge=False,
        show_footer=False,
        show_lines=True,
        show_header=True,
        expand=True,
        box=box.ROUNDED,
    )

    vizualisation_name = "Category Imbalance" if classification else "Histogram"
    vizualisation_func = distribution_plot if classification else histogram_plot
    table.add_column(
        "Series Type", justify="right", style="cyan", width=2, no_wrap=False
    )
    table.add_column(
        vizualisation_name, justify="left", style="cyan", width=10, no_wrap=False
    )
    table.add_column(
        "Types",
        justify="right",
        style="cyan",
        width=1,
        no_wrap=False,
    )
    table.add_column(
        "Indicies",
        justify="right",
        style="cyan",
        width=1,
        no_wrap=False,
    )

    table.add_row(
        "Targets",
        plotextMixin(y, vizualisation_func, title=vizualisation_name),
        f"NaNs: {str(calculate_nans(y))}\ndtype: {str(y.dtype)}",
        f"Start: {str(y.index[0])}\nEnd: {str(y.index[-1])}",
    )
    table.add_row(
        "Predictions",
        plotextMixin(preds, vizualisation_func, title=vizualisation_name),
        f"NaNs: {str(calculate_nans(preds))}\ndtype: {str(preds.dtype)}",
        f"Start: {str(preds.index[0])}\nEnd: {str(preds.index[-1])}",
    )
    return table



from typing import List

from krisi.evaluate.scorecard import ScoreCard
from krisi.utils.printing import bold


def compare(
    scorecards: List[ScoreCard],
    sort_metric_key: str = "rmse",
    metrics_to_display: List[str] = [],
) -> None:

    metric_title = "".join([f"{metric:<15s}" for metric in metrics_to_display])
    print(
        f"{'model_name':>30s}    {bold(f'{sort_metric_key:<15s}', rich=False)} {metric_title}"
    )
    scorecards.sort(reverse=True, key=lambda x: x[sort_metric_key].result)

    for scorecard in scorecards:
        metrics = "".join(
            [f"{scorecard[metric].result:<15.5}" for metric in metrics_to_display]
        )

        print(
            f"{scorecard.metadata.model_name:>30s}    {bold(f'{scorecard[sort_metric_key].result:<15.5}', rich=False)} {metrics}"
        )



import datetime
import uuid
from typing import Optional, Tuple

import pandas as pd

from krisi.evaluate.type import Targets


def handle_unnamed(
    y: Targets,
    model_name: Optional[str],
    dataset_name: Optional[str],
    project_name: Optional[str],
) -> Tuple[str, str, str]:

    time = datetime.datetime.now()
    display_time = time.strftime("%Y%m%d-%H%M%S")

    if dataset_name is None:
        if isinstance(y, pd.Series) and y.name is not None:
            dataset_name = str(y.name)
        else:
            dataset_name = f"Dataset:{display_time+str(uuid.uuid4()).split('-')[0]}"

    if model_name is None:
        model_name = f"Model:{display_time+str(uuid.uuid4()).split('-')[0]}"

    if project_name is None:
        project_name = f"Project:{display_time+str(uuid.uuid4()).split('-')[0]}"

    return model_name, dataset_name, project_name



import base64
import pkgutil
from typing import Dict, List

import plotly.graph_objects as go

from krisi.report.type import InteractiveFigure, PathConst


def __figure_to_html(figure: go.Figure) -> str:
    image = str(base64.b64encode(figure.to_image(format="png", scale=5)))[2:-1]
    return f'<img style="max-width:100%; max-height:100%;" src="data:image/png;base64,{image}"/>'


def __create_html_report(
    template_file: str, html_elements_to_inject: dict[str, str]
) -> str:
    template_html = pkgutil.get_data(__name__, template_file)
    template_html = template_html.decode()

    for key, value in html_elements_to_inject.items():
        template_html = template_html.replace(f"{{ {key} }}", value)
    return template_html


def __convert_html_to_pdf(
    source_html: str, output_path: str, report_name: str, css_template_url: str
) -> None:
    import os

    from weasyprint import CSS, HTML

    if not os.path.exists(output_path):
        os.makedirs(output_path)

    css_string = pkgutil.get_data(__name__, css_template_url)
    css_string = css_string.decode()

    css = CSS(string=css_string)
    pdf = HTML(string=source_html).write_pdf(stylesheets=[css])

    open(f"{output_path}/{report_name}", "wb").write(pdf)


def convert_figures_to_html(figures: List[go.Figure]) -> str:
    return "<br>".join([__figure_to_html(figure) for figure in figures])


def create_pdf_report(
    path: str = PathConst.default_save_path,
    title: str = "Time Series Report",
    html_template_url: str = PathConst.html_template_url,
    css_template_url: str = PathConst.css_template_url,
    html_elements_to_inject: dict[str, str] = dict(),
):
    html_elements_to_inject["title"] = title
    report_html = __create_html_report(html_template_url, html_elements_to_inject)
    __convert_html_to_pdf(report_html, path, f"{title}.pdf", css_template_url)


def append_sizes(
    diagram_dict: Dict[str, InteractiveFigure]
) -> Dict[str, InteractiveFigure]:

    size_dict = dict(
        residuals_display_acf_plot=(750.0, 750.0),
        residuals_display_density_plot=(350.0, 350.0),
        residuals_display_time_series=(1500.0, 750.0),
    )

    for key, value in size_dict.items():
        if key in diagram_dict.keys():
            diagram_dict[key].width = value[0]
            diagram_dict[key].height = value[1]

    return diagram_dict



import numpy as np

from krisi.evaluate import SampleTypes, ScoreCard

target, predictions = np.random.rand(1000), np.random.rand(1000)
sc = ScoreCard(
    target,
    predictions,
    model_name="<your_model_name>",
    dataset_name="<your_dataset_name>",
    sample_type=SampleTypes.insample,
)

""" A predefined metrics """
sc.evaluate(defaults=True)

""" Adding a new metric """
sc["own_metric"] = (target - predictions).mean()

""" Print scorecard summary """
sc.print_summary(extended=True)



import numpy as np

from krisi import score
from krisi.evaluate.type import SaveModes

scorecard = score(y=np.random.rand(1000), predictions=np.random.rand(1000))
scorecard.save(
    save_modes=[
        SaveModes.minimal,
        SaveModes.obj,
        SaveModes.text,
        SaveModes.svg,
        SaveModes.html,
        SaveModes.text,
    ]  # Optional, defaults to [ SaveModes.minimal, SaveModes.obj, SaveModes.text ]
)



import pandas as pd
import plotext as plx


def histogram_plot(data: pd.Series, width: float, height: float, title: str) -> str:
    plx.clf()
    plx.hist(data, 40)
    plx.plotsize(width, 10)
    plx.theme("dark")

    return plx.build()


def line_plot(data: pd.Series, width: float, height: float, title: str) -> str:
    plx.clf()
    plx.plot(data, marker="hd")
    plx.plotsize(width, 10)
    plx.theme("dark")

    return plx.build()


def distribution_plot(data: pd.Series, width: float, height: float, title: str) -> str:
    plx.clf()
    data_ = data.to_frame("values")

    groups = data_.groupby("values").size()
    plx.simple_bar(
        groups.index,
        groups.to_list(),
        width=width,
    )

    return plx.build()



import numpy as np

from krisi import compare, score

scorecards = [
    score(y=np.random.randint(0, 2, 1000), predictions=np.random.randint(0, 2, 1000))
    for _ in range(5)
]
compare(
    scorecards, sort_metric_key="accuracy", metrics_to_display=["recall", "f_one_score"]
)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Tuple, Union

from statsmodels.tsa.arima.model import ARIMA

from krisi.utils.modeling_types import InSamplePredictions, OutSamplePredictions, X, y


class Model(ABC):

    name: str = ""

    @abstractmethod
    def fit(self, X: X, y: y) -> None:
        raise NotImplementedError

    @abstractmethod
    def predict(self, X: X) -> OutSamplePredictions:
        raise NotImplementedError

    @abstractmethod
    def predict_in_sample(self, X: X) -> InSamplePredictions:
        raise NotImplementedError


class StrategyTypes(Enum):
    mean = "mean"
    last = "last"
    drift = "drift"


@dataclass
class NaiveForecasterConfig:
    strategy: StrategyTypes
    fh: Optional[Union[int, List[int]]] = None  # Forecasting Horizon
    window_length: Optional[int] = None  # The window of the mean if strategy is 'mean'
    sp: int = 1  # Seasonal periodicity


@dataclass
class ArimaConfig:
    """
    order=(p,d,q).
        p is the autoregressive order
        d is the number of times the series has been differenced
        q is the moving average order
    """

    order: Tuple[float, float, float]  # Forecasting Horizon
    trend: Optional[Union[str, Tuple[float]]] = None


class ArimaWrapper(Model):

    name: str = ""

    def __init__(self, config: ArimaConfig) -> None:
        self.config = config

    def fit(self, X: X, y: y) -> None:
        unfitted_model = ARIMA(X, order=self.config.order, trend=self.config.trend)
        self.model = unfitted_model.fit()

    def predict_in_sample(self, X: X) -> InSamplePredictions:
        return self.model.predict(end=len(X) - 1)

    def predict(self, X: X) -> OutSamplePredictions:
        if self.model:
            return self.model.forecast(len(X))
        else:
            raise ValueError("Model has to be fitted first")


default_arima_model = ArimaWrapper(ArimaConfig(order=(1, 0, 1)))



import os

import numpy as np

from krisi.evaluate import SampleTypes, ScoreCard
from krisi.report.type import DisplayModes

target, predictions = np.random.rand(100), np.random.rand(100)
sc = ScoreCard(
    target,
    predictions,
    model_name="ARIMA",
    model_description="You can explain main features of the project to be displayed in the report",
    dataset_name="APPLE Stocks",
    dataset_description="You can explain main features of the dataset to be displayed in the report",
    project_name="Quantitative Finance Project",
    project_description="The extra information about what the project is about will be displayed when creating a report",
    sample_type=SampleTypes.insample,
)

""" A predefined metrics """
sc.evaluate(defaults=True)

if "PYTEST_CURRENT_TEST" in os.environ:
    print("Not testing Dash server currently")
else:
    sc.generate_report(
        display_modes=[DisplayModes.interactive],
    )



from typing import Union

import numpy as np
import pandas as pd

ModelOverTime = pd.Series
InSamplePredictions = Union[np.ndarray, pd.Series]
OutSamplePredictions = Union[np.ndarray, pd.Series]
y = Union[np.ndarray, pd.Series]
X = Union[np.ndarray, pd.DataFrame]
